#!/usr/bin/env python
from avi.sdk.avi_api import ApiSession
import uuid
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# Avi class
class Avi():

    def __init__(self, **params):
        self.controller = params.get('controller')
        self.username = params.get('username')
        self.password = params.get('password')
        self.tenant = params.get('tenant') if params.get('tenant') else "admin"
        self.api_version = params.get('api_version') if params.get('api_version') else "17.2.4"
        self.api = self.login()

    def login(self):
        return ApiSession.get_session(self.controller, self.username, self.password, tenant=self.tenant, api_version=self.api_version)

    def logout(self):
        return self.api.delete_session()

    # Get methods
    def get_objs(self, obj='cluster', name=None, tenant='admin'):
        if not name:
            res = self.api.get(obj, tenant).json()
            if "results" in res:
                return res['results']
            else:
                return res
        else:
            return self.api.get_object_by_name(obj, name, tenant)

    #
    # Virtual Service
    # 

    def get_version(self, name=None, tenant="admin"):
        return self.api.get('initial-data', tenant).json()['version']

    def get_virtualservices(self, name=None, tenant="admin"):
        params = {}
        params['include_name'] = True
        if not name:
            return self.api.get('virtualservice', tenant, params=params).json()['results']
        else:
            return self.api.get_object_by_name('virtualservice', name, tenant, params=params)
    
    def get_virtualservices_inventory(self, name=None, tenant="admin"):
        params = {}
        params['include_name'] = True
        if not name:
            return self.api.get('virtualservice-inventory', tenant, params=params).json()['results']
        else:
            return self.api.get_object_by_name('virtualservice-inventory', name, tenant, params=params)
    
    def disable_virtualservice(self, name=None, tenant="admin"):
        payload = self.get_virtualservices(name, tenant)
        payload["enabled"] = False
        return self.api.put_by_name('virtualservice', name, data=payload, tenant=tenant)
    
    def enable_virtualservice(self, name=None, tenant="admin"):
        payload = self.get_virtualservices(name, tenant)
        payload["enabled"] = True
        return self.api.put_by_name('virtualservice', name, data=payload, tenant=tenant)
    
    def scaleout_virtualservice(self, name=None, tenant="admin"):
        payload = self.get_virtualservices(name, tenant)
        uuid = payload["uuid"]
        path = "virtualservice/" + uuid + "/scaleout"
        data = {}
        data['vip_id'] = "0" 
        return self.api.post(path, data=data, tenant=tenant)
    
    def scalein_virtualservice(self, name=None, tenant="admin"):
        payload = self.get_virtualservices(name, tenant)
        uuid = payload["uuid"]
        path = "virtualservice/" + uuid + "/scalein"
        data = {}
        data['vip_id'] = "0" 
        return self.api.post(path, data=data, tenant=tenant)
    
    def get_pools(self, name=None, tenant="admin"):
        if not name:
            return self.api.get('pool', tenant).json()['results']
        else:
            return self.api.get_object_by_name('pool', name, tenant)

    def get_pools_inventory(self, name=None, tenant="admin"):
        if not name:
            return self.api.get('pool-inventory', tenant).json()['results']
        else:
            return self.api.get_object_by_name('pool-inventory', name, tenant)
    
    def disable_pool(self, name=None, tenant="admin"):
        payload = self.get_pools(name, tenant)
        payload["enabled"] = False
        return self.api.put_by_name('pool', name, data=payload, tenant=tenant)
    
    def enable_pool(self, name=None, tenant="admin"):
        payload = self.get_pools(name, tenant)
        payload["enabled"] = True
        return self.api.put_by_name('pool', name, data=payload, tenant=tenant)
    
    def disable_pool_member(self, member=None, name=None, tenant="admin"):
        if member != None and name != None:
            print(name)
            print(member)
            print(tenant)
            payload = self.get_pools(name, tenant)
            print(payload)
            for s in payload["servers"]:
                if member == s["hostname"] or member == s["ip"]["addr"]:
                    s["enabled"] = False
            return self.api.put_by_name('pool', name, data=payload, tenant=tenant)
    
    def enable_pool_member(self, member=None, name=None, tenant="admin"):
        if member != None and name != None:
            payload = self.get_pools(name, tenant)
            for s in payload["servers"]:
                if member == s["hostname"] or member == s["ip"]["addr"]:
                    s["enabled"] = True
            return self.api.put_by_name('pool', name, data=payload, tenant=tenant)
    
    
    def get_serviceengines(self, name=None, tenant="admin"):
        if not name:
            return self.api.get('serviceengine', tenant=tenant).json()['results']
        else:
            return self.api.get_object_by_name('serviceengine', name)
    
    def get_serviceenginegroups(self, name=None, tenant="admin"):
        if not name:
            return self.api.get('serviceenginegroup', tenant).json()['results']
        else:
            return self.api.get_object_by_name('serviceenginegroup', name, tenant)
    
    # Create methods
    def create_virtualservice(self, **obj):
        try:

            # Set defaults
            ssl_certs = "System-Default-Cert"
            ssl_profile = "System-Standard"
            app_profile = "System-HTTP"
            network_profile = "System-TCP-Proxy"
            lb_algo = "LB_ALGORITHM_LEAST_CONNECTIONS"
            global_ssl_flag = False

            # Payload
            payload = {}

            # Use standard object (vs. macro)
            
            # Virtual Service name
            if obj.get('name'):
                payload["name"] = obj.get('name')
            else:
                payload["name"] = "vs-" + str(uuid.uuid4()).split("-")[4]

            # Virtual Service cloud, will default to "Default-Cloud"
            if obj.get('cloud'):
                payload["cloud_ref"] = "https://" + self.api.controller_ip + "/api/cloud?name=" + obj.get('cloud')
            else:
                payload["cloud_ref"] = "https://" + self.api.controller_ip + "/api/cloud?name=Default-Cloud"
            
            # Build the services
            if obj.get('ports'):
                payload["services"] = []
                for port in obj.get('ports').split(","):
                    tmp = {}
                    if ":" in port:
                        port_l4, enable_ssl = port.split(":")
                        if enable_ssl:
                            tmp["port"] = port_l4
                            if enable_ssl == "SSL" or enable_ssl == "ssl":
                                tmp["enable_ssl"] = True
                            global_ssl_flag = True
                    else:
                        tmp["port"] = port
                    payload["services"].append(tmp)
                    
            # Build the pool ref
            payload["pool_ref"] = {}
            if obj.get('pool'):
                payload["pool_ref"] = "https://" + self.api.controller_ip + "/api/pool?name=" + obj.get('pool')
            
            # Enable RHI
            if obj.get('bgp'):
                payload["enable_rhi"] = True

            # Currently VIP requires IPAM profile for this module
            payload["vip"] = []
            vip = {}
            vip["auto_allocate_ip"] = True
            vip["ipam_network_subnet"] = {}
            vip["ipam_network_subnet"]["network_ref"] = "https://" + self.api.controller_ip + "/api/network?name=" + obj.get('network')
            payload["vip"].append(vip)
            
            # SE Group, if not Default-Group
            if obj.get('se_group'):
                payload["se_group_ref"] = "https://" + self.api.controller_ip + "/api/serviceenginegroup?name=" + obj.get('se_group')
                
            # Post the payload
            resp = self.api.post('virtualservice', data=payload)
            if resp.status_code == 201:
                print(resp.text)
                return resp.json()
            else:
                print(resp.text)
                return resp

        except:
            print("Error with POST to /api/virtualservice")

    def create_pool(self, **obj):
        try:
            payload = {}

            # Pool name
            if obj.get('name'):
                payload["name"] = obj.get('name')
            else:
                payload["name"] = "pool-" + str(uuid.uuid4()).split("-")[4]

            # Pool cloud, will default to "Default-Cloud"
            if obj.get('cloud'):
                payload["cloud_ref"] = "https://" + self.api.controller_ip + "/api/cloud?name=" + obj.get('cloud')
            
            # Pool LB algo
            if obj.get('lb_algo'):
                payload["lb_algorithm"] = obj.get('lb_algo')
			
            # Pool Servers, will default to empty
            payload["servers"] = []
            if obj.get('servers'):
                for ipa in obj.get('servers').split(","):
                    tmp = {}
                    if ":" in ipa:
                        ip, port = ipa.split(":")
                        payload["default_server_port"] = port
                        tmp["port"] = port
                    else:
                        ip = ipa
                        tmp["ip"] = {}
                        tmp["ip"]["type"] = "V4"
                        tmp["ip"]["addr"] = ip
                        payload["servers"].append(tmp)
        
            # Post the payload
            resp = self.api.post('pool', data=payload)
            if resp.status_code == 201:
                print(resp.text)
                return resp.json()
            else:
                print(resp.text)
                return resp

        except:
            print("Error with POST to /api/pool")
