#!/usr/bin/env python
#
# Updated on Mar 1, 2018
# @author: Tory Long (tory@avinetworks.com)
#
# AviBot - Slack bot for Avi integration
#

import os
import time
from slack import WebClient, RTMClient
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import json
import re
from avi_mod import Avi
import threading

# Suppress warnings for now
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

def invoke_slack_client(token):
    """ invoke_slack_client
    Returns SlackClient instance given token
    params:
        token - SlackClient token in form beginning with xoxo
    """
    # Configure the SlackClient w/ ID
    try:
        return WebClient(token)
    except:
        print("Unable to get SlackClient instance.")

def invoke_avi_session(controller, username, password, tenant=None, api_version=None):
    """ invoke_avi_session
    Returns Avi session instance given token
    params:
        controller - Avi controller IP/FQDN
        username - Avi controller username
        password - Avi controller password
        tenant - (optional) Avi tenant
        api_version - (optional) Avi API version
    """
    # Get an Avi session
    try:
        return Avi(controller=controller, username=username, password=password, tenant=tenant, api_version=api_version)
    except:
        print("Unable to get Avi session.")

def add_avi_footer(session):
    """ add_avi_footer
    Returns dict for footer inclusion on Slack message
    params:
        session - Avi session
    """

    # Dict to hold footer
    tmp_att = {}

    # Build footer
    controller_ip = session.api.controller_ip
    controller_string = "Controller: " + controller_ip
    controller_url = "https://" + controller_ip
    tmp_att["title_link"] = controller_url
    tmp_att["footer_icon"] = controller_url + "/src/img/favicon.ico"
    tmp_att["footer"] = controller_string

    # Return merged footer
    return tmp_att

def add_avi_title_link(session, vip=None):
    """ add_avi_title_link
    Returns dict for title inclusion inclusion on Slack message
    params:
        session - Avi session
        vip - (optional) Controller VIP if using FQDN in title
    """
    
    # Dict to hold footer
    tmp_att = {}

    # Build title
    controller_ip = session.api.controller_ip
    controller_url = "https://" + controller_ip
    tmp_att["title"] = controller_ip
    if vip:
        tmp_att["title"] += " (" + vip + ")"
    tmp_att["title_link"] = controller_url
    
    # Return merged footer
    return tmp_att


def handle_slack_command(command, channel, user=None, slack_web_client=None):
    """ handle_slack_command
    Handles individual Slack commands towards bot
    params:
        command - Slack command following parsing
        channel - Slack channel
        user - Slack user who typed query to bot
        slack_web_client - slack session
    """

    # connect to Avi, now that we have a valid command
    session = invoke_avi_session(
        controller=os.environ.get('AVI_CONTROLLER'),
        username=os.environ.get('AVI_USERNAME'),
        password=os.environ.get('AVI_PASSWORD'),
        tenant=os.environ.get('AVI_TENANT'),
        api_version=os.environ.get('AVI_API_VERSION')
        )


    # connect to slack webapi
    user_info = slack_web_client.users_info(user=user)
    #print(user_info)
    user_real_name = user_info['user']['real_name']

    # Create some vars
    response = ""
    attachments = []

    # Parse the command
    comm_arr = command.split(" ")
    comm_arr[0] = comm_arr[0].lower()

    # Construct the action/response for specific commands
    
    # hello
    if comm_arr[0] == "hello":
        json_data = session.get_objs()
        response = f"Hi {user_real_name}, you're connected to Avi!"
        if len(json_data["uuid"]) > 0:
            tmp_att = {}
            if "virtual_ip" in json_data:
                vip = json_data["virtual_ip"]["addr"] or json_data["nodes"][0]["ip"]["addr"]
                tmp_att.update(add_avi_title_link(session, vip))
            else:
                tmp_att.update(add_avi_title_link(session))
            name = json_data["name"]
            uuid = json_data["uuid"]
            tmp_att["color"] = "#008000"
            tmp_att.update(add_avi_footer(session))
        attachments.append(tmp_att)
    
    # scalein and scaleout
    elif "scale" in comm_arr[0]:
        if comm_arr[1] == "vs":
            if len(comm_arr) == 3:
                if "out" in comm_arr[0]:
                    json_data = session.scaleout_virtualservice(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
                    response = f"OK {user_real_name}, scaling out virtual service *" + comm_arr[2] + "*"
                else:
                    json_data = session.scalein_virtualservice(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
                    response = f"OK {user_real_name}, scaling in virtual service *" + comm_arr[2] + "*"
                time.sleep(0.5)
        else:
            response = "This is where we'll implement some Avi creation"
        comm_arr = ["show","vs",comm_arr[2],"detail"]
    
    # enable and disable
    elif comm_arr[0] == "disable" or comm_arr[0] == "enable":
        if comm_arr[1] == "pool":
            if len(comm_arr) >= 3:
                if len(comm_arr) == 5:
                    if comm_arr[0] == "disable":
                        json_data = session.disable_pool_member(member=comm_arr[4], name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
                        response = f"OK {user_real_name}, disabling pool member *" + comm_arr[4] + "*"
                    else: 
                        json_data = session.enable_pool_member(member=comm_arr[4], name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
                        response = f"OK {user_real_name}, enabling pool member *" + comm_arr[4] + "*"
                else:
                    if comm_arr[0] == "disable":
                        json_data = session.disable_pool(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
                    else: 
                        json_data = session.enable_pool(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
            time.sleep(0.6)
            comm_arr = ["show","pool",comm_arr[2],"detail"]
        elif comm_arr[1] == "vs":
            if len(comm_arr) >= 3:
                if len(comm_arr) == 3:
                    if comm_arr[0] == "disable":
                        json_data = session.disable_virtualservice(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
                    else: 
                        json_data = session.enable_virtualservice(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT'))
            time.sleep(0.6)
            comm_arr = ["show","vs",comm_arr[2],"detail"]
    
    # show or get
    elif comm_arr[0] == "show" or comm_arr[0] == "get":
        if "pool" in comm_arr[1]:
            if len(comm_arr) >= 2:
                if len(comm_arr) == 2 or "detail" in comm_arr[2]:
                    json_data = session.get_pools_inventory(tenant=os.environ.get('AVI_TENANT'))
                    response = f"OK {user_real_name}, showing all pools"
                else:
                    json_data = []
                    json_data.append(session.get_pools_inventory(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT')))
                    if json_data == [None]:
                        json_data = session.get_pools_inventory(tenant=os.environ.get('AVI_TENANT'))
                        if json_data != [None]:
                            regex = r"" + comm_arr[2].replace("-","") + ".*"
                            json_data = filter(lambda c: re.search(regex, c["config"]["name"]), json_data)
                    if len(response) < 1:
                        response = f"OK {user_real_name}, showing pools matching *" + comm_arr[2] + "*"
                json_data = [] if json_data == [None] else json_data

                for o in json_data:
                    if len(o["config"]["name"]) > 0:
                        tmp_att = {}

                        if o["runtime"]["num_servers_up"] == o["runtime"]["num_servers"] and o["config"]["enabled"]:
                            tmp_att["color"] = "#008000"
                        elif o["runtime"]["num_servers_up"] == 0 and o["runtime"]["num_servers"] != 0 and o["config"]["enabled"]:
                            tmp_att["color"] = "#FF0000"
                        elif not o["config"]["enabled"]:
                            tmp_att["color"] = "#7C7C7C"
                        else: 
                            tmp_att["color"] = "#F8F800"

                        # Title link
                        tmp_att["title"] = o["config"]["name"]
                        tmp_att["title"] += "  (" + str(o["runtime"]["num_servers_up"]) + "/"
                        tmp_att["title"] += str(o["runtime"]["num_servers"]) + ")"
                        tmp_att["title_link"] = o["config"]["url"].replace("/api/", "/#!/applications/")
                        if len(o["virtualservices"]) > 0:
                            tmp_att["title_link"] += o["virtualservices"][0].replace("/api/virtualservice", "")
                            tmp_att["title_link"] += "/analytics"
                        
                        # Add details
                        if "detail" in comm_arr or "details" in comm_arr:
                            d_json_data = []
                            d_json_data.append(session.get_pools(name=o["config"]["name"], tenant=os.environ.get('AVI_TENANT')))
                            for d in d_json_data:
                                tmp_att["fields"] = []
                            
                                tmp_field = {}
                                tmp_field["title"] = "Servers"
                                tmp_field["value"] = ""
                                if "servers" in d:
                                    for svr in d["servers"]:
                                        if "hostname" in svr and svr["hostname"] != "":
                                            tmp_field["value"] += svr["hostname"] + " ("
                                            tmp_field["value"] += svr["ip"]["addr"] + ")"
                                        else:
                                            tmp_field["value"] += svr["ip"]["addr"]
                                        tmp_field["value"] += "\n"
                                tmp_field["short"] = True
                                tmp_att["fields"].append(tmp_field)
                            
                                tmp_field = {}
                                tmp_field["title"] = "State"
                                tmp_field["value"] = ""
                                if "servers" in d:
                                    for svr in d["servers"]:
                                        if svr["enabled"]:
                                            tmp_field["value"] += "Enabled"
                                        else:
                                            tmp_field["value"] += "Disabled"
                                        tmp_field["value"] += "\n"
                                tmp_field["short"] = True
                                tmp_att["fields"].append(tmp_field)
                                tmp_att.update(add_avi_footer(session))
                        attachments.append(tmp_att)
    
        elif comm_arr[1] == "vs":
            if len(comm_arr) >= 2:
                if len(comm_arr) == 2 or "detail" in comm_arr[2] or "health" in comm_arr[2]:
                    json_data = session.get_virtualservices_inventory(tenant=os.environ.get('AVI_TENANT'))
                    response = f"OK {user_real_name}, showing all virtual services"
                else:
                    json_data = []
                    json_data.append(session.get_virtualservices_inventory(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT')))
                    if json_data == [None]:
                        json_data = session.get_virtualservices_inventory(tenant=os.environ.get('AVI_TENANT'))
                        if json_data != [None]:
                            regex = r"" + comm_arr[2].replace("-","") + ".*"
                            json_data = filter(lambda c: re.search(regex, c["config"]["name"]), json_data)
                    if len(response) < 1:
                        response = f"OK {user_real_name}, showing virtual services matching *" + comm_arr[2] + "*"
                json_data = [] if json_data == [None] else json_data
                for o in json_data:
                    if len(o["pools"]) > 0:
                        tmp_att = {}
                        if o["config"]["enabled"] and o["runtime"]["vip_summary"][0]["num_se_assigned"] > 0:
                            if o["runtime"]["vip_summary"][0]["percent_ses_up"] == 100:
                                tmp_att["color"] = "#008000"
                            elif o["runtime"]["vip_summary"][0]["percent_ses_up"] < 100 and o["runtime"]["oper_status"]["state"] == "OPER_UP":
                                tmp_att["color"] = "#F8F800"
                            else:
                                tmp_att["color"] = "#FF0000"
                        elif not o["config"]["enabled"]:
                            tmp_att["color"] = "#7C7C7C"
                        else: 
                            tmp_att["color"] = "#FF0000"

                        tmp_att["title"] = o["config"]["name"]
                        if "dns_info" in o["config"] and len(o["config"]["dns_info"]) > 0:
                            tmp_att["title"] += " (" + o["config"]["dns_info"][0]["fqdn"] + ")"
                        tmp_att["title_link"] = o["config"]["url"].replace("/api/","/#!/applications/")
                        tmp_att["title_link"] += "/analytics"
			
                        if "detail" in comm_arr or "details" in comm_arr:
                            tmp_att["text"] = "Pools: " 
                            for x in range(0, len(o["pools"])):
                                pool_uuid = o["pools"][x].replace("/api/pool/", "")
                                p_uuid_short = "-".join(pool_uuid.split("-")[0:2])
                                tmp_att["text"] += p_uuid_short
                                if x != len(o["pools"]) - 1:
                                    tmp_att["text"] += ", "

                            d_json_data = []
                            d_json_data.append(session.get_virtualservices(name=o["config"]["name"], tenant=os.environ.get('AVI_TENANT')))
                            
                            for d in d_json_data:
                                tmp_att["fields"] = []

                                # First field
                                tmp_field = {}
                                tmp_field["title"] = "VIPs"
                                tmp_field["value"] = ""
                                if "vip" in d:
                                    for vip in d["vip"]:
                                        tmp_field["value"] += vip["ip_address"]["addr"]
                                tmp_field["short"] = True
                                tmp_att["fields"].append(tmp_field)

                                # Second field 
                                tmp_field = {}
                                tmp_field["title"] = "Services"
                                tmp_field["value"] = ""
                                tmp_count = 0
                                if "services" in d:
                                    for svc in d["services"]:
                                        tmp_count += 1
                                        if svc["port"] == svc["port_range_end"]:
                                            tmp_field["value"] += str(svc["port"])
                                        else:
                                            tmp_field["value"] += str(svc["port"])+"-"+str(svc["port_range_end"])
                                        if tmp_count < len(d["services"]):
                                            tmp_field["value"] += ", "
                                tmp_field["short"] = True
                                tmp_att["fields"].append(tmp_field)
				
                                # Third field 
                                tmp_field = {}
                                tmp_field["title"] = "Service Engine Group"
                                tmp_field["value"] = d['se_group_ref'].split("#")[1]
                                tmp_field["short"] = True
                                tmp_att["fields"].append(tmp_field)
                                
                                # Fourth field 
                                tmp_field = {}
                                tmp_field["title"] = "Service Engines"
                                tmp_field["value"] = ""
                                tmp_count = 0
                                if "vip_runtime" in d:
                                    if "se_list" in d["vip_runtime"][0]:
                                        for se in d["vip_runtime"][0]["se_list"]:
                                            tmp_count += 1
                                            tmp_field["value"] += se["se_ref"].split("#")[1]
                                            if tmp_count < len(d["vip_runtime"][0]["se_list"]):
                                                tmp_field["value"] += "\n"
                                tmp_field["short"] = True
                                tmp_att["fields"].append(tmp_field)
                                
                                # Append all fields
                                tmp_att.update(add_avi_footer(session))
                        
                        elif "health" in comm_arr:
                            tmp_att["fields"] = []
                            
                            metric_names = {}
                            metric_names["title"] = "Health Metric"
                            metric_names["value"] = ""
                            metric_names["short"] = True
                            
                            metric_scores = {}
                            metric_scores["title"] = "Score"
                            metric_scores["value"] = ""
                            metric_scores["short"] = True
                            
                            overall_health_score = "0.0"

                            if "health_score" in o:
                                metric_names["value"] += "Performance\n"
                                metric_names["value"] += "Resources Penalty\n"
                                metric_names["value"] += "Anomaly Penalty\n"
                                metric_names["value"] += "Security Penalty\n"
                                metric_names["value"] += "Overall Health\n"
                                
                                metric_scores["value"] += str(int(o["health_score"]["performance_score"])) + "\n"
                                metric_scores["value"] += str(0 - int(o["health_score"]["resources_penalty"])) + "\n"
                                metric_scores["value"] += str(0 - int(o["health_score"]["anomaly_penalty"])) + "\n"
                                metric_scores["value"] += str(0 - int(o["health_score"]["security_penalty"])) + "\n"
                                metric_scores["value"] += str(int(o["health_score"]["health_score"])) + "\n"
                                
                                score = int(o["health_score"]["health_score"])
						
                                if score < 80 and score > 0:
                                    tmp_att["color"] = "#7C7C7C"
                                elif score > 80:
                                    tmp_att["color"] = "#FF0000"
                                else:
                                    tmp_att["color"] = "#F8F800"

                            tmp_att["fields"].append(metric_names)
                            tmp_att["fields"].append(metric_scores)
                                
                            tmp_att.update(add_avi_footer(session))

                        attachments.append(tmp_att)
            else:
                response = "show command: Not implemented"
        elif comm_arr[1] == "se":
            if len(comm_arr) >= 2:
                if len(comm_arr) == 2 or comm_arr[2] == "detail":
                    json_data = session.get_serviceengines(tenant=os.environ.get('AVI_TENANT'))
                    response = f"OK {user_real_name}, showing all service engines"
                else:
                    json_data = []
                    json_data.append(session.get_serviceengines(name=comm_arr[2], tenant=os.environ.get('AVI_TENANT')))
                    response = f"OK {user_real_name}, showing service engine *" + comm_arr[2] + "*"
                json_data = [] if json_data == [None] else json_data
                for o in json_data:
                    status = o["state_name"].split("::")[1]
                    if status in ["Online"]:
                        controller_ip = o["url"].split("/")[2]
                        controller_string = "Controller: " + controller_ip
                        controller_url = "https://" + controller_ip
                        
                        tmp_att = {}
                        tmp_att["color"] = "#008000" if status == "Online" else "#FF0000"
                        tmp_att["footer"] = controller_string
                        tmp_att["title"] = o["name"]
                        mgmt_ip = o["mgmt_vnic"]["vnic_networks"][0]["ip"]["ip_addr"]["addr"]
                        tmp_att["title"] += " (" + mgmt_ip + ")"
                        tmp_att["title_link"] = re.sub("api/cloud", "#/infrastructure/serviceengine", o["cloud_ref"])
                        tmp_att["text"] = "Status: " + o["oper_status"]["state"]
                        tmp_att["text"] += ", " + o["oper_status"]["reason"][0]
	
                        tmp_att["fields"] = []
	
                        tmp_field = {}
                        tmp_field["title"] = "Operational Status"
                        tmp_field["value"] = o["oper_status"]["state"]
                        tmp_field["short"] = True
                        tmp_att["fields"].append(tmp_field)
                        
                        tmp_field = {}
                        tmp_field["title"] = "Reason"
                        tmp_field["value"] = o["oper_status"]["reason"][0]
                        tmp_field["short"] = True
                        tmp_att["fields"].append(tmp_field)
	
                        tmp_att["footer_icon"] = controller_url + "/src/img/favicon.ico"
                        tmp_att["ts"] = o["oper_status"]["last_changed_time"]["secs"]
                        
                        attachments.append(tmp_att)

        elif comm_arr[1] == "version":
            json_data = session.get_version()
            response = f"OK {user_real_name}, showing version for the Avi cluster"
            if len(json_data["Version"]) > 0:
                tmp_att = {}
                tmp_att.update(add_avi_title_link(session))
                tmp_att.update(add_avi_footer(session))
                tmp_att["color"] = "#008000"
                tmp_att["fields"] = [] 
                tmp_field = {}
                tmp_field["title"] = "Version"
                tmp_field["value"] = ""
                tmp_field["value"] += json_data["Version"] + " ("
                tmp_field["value"] += str(json_data["build"]) + ")\n"
                tmp_att["fields"].append(tmp_field)
            attachments.append(tmp_att)
        
        elif comm_arr[1] == "cluster":
            json_data = session.get_objs()
            response = f"OK {user_real_name}, showing details for the Avi cluster"
            if len(json_data["uuid"]) > 0:
                tmp_att = {}
                if "virtual_ip" in json_data:
                    vip = json_data["virtual_ip"]["addr"] or json_data["nodes"][0]["ip"]["addr"]
                    tmp_att.update(add_avi_title_link(session, vip))
                else:
                    tmp_att.update(add_avi_title_link(session))
                tmp_att.update(add_avi_footer(session))
                tmp_att["color"] = "#008000"
                tmp_att["fields"] = []
                tmp_field = {}
                tmp_field["title"] = "Nodes"
                tmp_field["value"] = ""
                for n in json_data["nodes"]:
                    if "vm_hostname" in n and n["vm_hostname"] != "":
                        tmp_field["value"] += n["vm_hostname"] + " ("
                        tmp_field["value"] += n["ip"]["addr"] + ")\n"
                tmp_att["fields"].append(tmp_field)
                
            attachments.append(tmp_att)

        else:
            response = "Subcommand not found, supported commands are:\n•pool\n•vs\n•se\n•version\n•cluster"
	
    else:
        response = "You should not have gotten here, this is an error, big-style"

    if len(attachments) > 0:
        # Post the attachments
        slack_web_client.chat_postMessage(channel=channel, text=response, attachments=attachments, as_user=True)
    else:
        # Handle non-implemented commands
        #response = "Avi object not found or command not implemented."
        slack_web_client.chat_postMessage(channel=channel, text=response, as_user=True)


@RTMClient.run_on(event="message")
def parse_slack_output(**payload):
    """ parse_slack_output
    Parses Slack real-time-messaging output
    params:
        slack_rtm_output - Slack real-time output
    """
    
    # Restrict commands we can handle
    accepted_commands = ["hello","show","get","disable","enable","scalein","scale-in","scaleout","scale-out"]
    slack_rtm_output = payload['data']

    #to debug realtime messages, enable the following print
    #print(slack_rtm_output)

    # Absorb the RTM output
    if slack_rtm_output['client_msg_id'] and len(slack_rtm_output) > 0:
        slack_web_client = invoke_slack_client(token=os.environ.get('SLACK_BOT_TOKEN'))

        command = slack_rtm_output['text'].strip()
        if command.split(" ")[0] in accepted_commands:
            print("Command", command, " Channel", slack_rtm_output['channel'], " User", slack_rtm_output['user'])
            handle_slack_command(command, slack_rtm_output['channel'], slack_rtm_output['user'],slack_web_client)

        else:
            print("command not supported•")
            slack_web_client = invoke_slack_client(token=os.environ.get('SLACK_BOT_TOKEN'))
            response = "Command not found, supported commands are:\n•hello\n•show\n•get\n•disable\n•enable\n•scalein\n•scale-in\n•scaleout\n•scale-out"
            slack_web_client.chat_postMessage(channel=slack_rtm_output['channel'], text=response, as_user=True)
    
    else:
        print("probably a bot message, not a user message")


if __name__ == "__main__":

    # Set the Bot name, fairly irrelevant
    BOT_NAME = "avibot-nick"

    # Get Slack client
    rtm_client = RTMClient(token=os.environ.get('SLACK_BOT_TOKEN'))
    rtm_client.start()

